#!/usr/bin/env python

## Copyright (C) 2002   Martin Willemoes Hansen

## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

## Author:   Martin Willemoes Hansen
## Version:  0.12.1

import sys, random
import pygame
from pygame.locals import *
from common import *

# constants
SCREENRECT = Rect (25, 25, 590, 455)

# classes
class Bar (pygame.sprite.Sprite):
    """A bar moving left and right on the bottom of the screen"""
    GLUE_TIME = 30000
    LEFT = 0
    RIGHT = 1

    def __init__ (self):
        pygame.sprite.Sprite.__init__ (self)
        self.normalize()
        self.updateGroup = pygame.sprite.RenderClear()
        self.add (self.updateGroup)

    def normalize (self):
        self.speed = 2
        self.glue = 0
        self.shots = 0
        self.reloading = 0
        self.timeGotGlue = 0
        self.timeGotDrawed = 0
        self.loadImage ('bar1')
        self.rect.bottom = (SCREENRECT.bottom - 10)
        self.rect.centerx = SCREENRECT.centerx

    def setGlue (self):
        self.glue = 1
        self.timeGotGlue = pygame.time.get_ticks()

    def draw (self):
        self.updateGroup.clear (screen.display, background)
        self.updateGroup.draw (screen.display)

    def move (self, direction):
        "Moves the amount to either right of left side"

        # Checks if glue has timed out and should be disabled
        if self.glue and pygame.time.get_ticks() > self.timeGotGlue + self.GLUE_TIME:
            self.glue = 0

        if direction == self.LEFT:
            self.rect.move_ip (-self.speed, 0)
        elif direction == self.RIGHT:
            self.rect.move_ip (self.speed, 0)

        self.rect = self.rect.clamp (SCREENRECT)

        self.draw()

    def loadImage (self, image):
        self.image = gfx[image]
        self.rect = self.image.get_rect()
        self.WIDTH = self.rect.width

    def setSpeed (self, amount):
        if self.speed >= 1:
            self.speed += amount

class Ball (pygame.sprite.Sprite):
    """A ball which can travel around the sourface, if it hits the bottom its lost"""
    ETHEREAL_TIME = 20000
    FLOOR_TIME = 10000
    ethereal = 0
    floor = 0
    speed = 0

    launched = 1
    diffXpos = 0

    def __init__ (self):
        pygame.sprite.Sprite.__init__ (self)

        self.image = gfx['ball']
        self.rect = self.image.get_rect()
        self.WIDTH = self.rect.width
        self.rect.centerx = bar.rect.centerx
        self.rect.centery = bar.rect.centery - self.WIDTH + 1

        self.setSpeed (1)
        self.direction = random.choice ((self.ne, self.nw))

    def setSpeed (self, amount):
        if (amount > 0 or self.speed > 1):
            self.speed += amount

            self.ne = (self.speed, -self.speed)
            self.nw = (-self.speed, -self.speed)
            self.se = (self.speed, self.speed)
            self.sw = (-self.speed, self.speed)

    def glue (self):
        self.rect.centerx = bar.rect.centerx + self.diffXpos

    def setEthereal (self):
        self.etherealTime = pygame.time.get_ticks()
        self.ethereal = 1

    def setFloor (self):
        self.floorTime = pygame.time.get_ticks()
        self.floor = 1

    def launch (self, boolean):
        if boolean:
            self.launched = 1
        else:
            self.launched = 0

    def update (self):
        "Moves the ball around acording to its direction and changes it too if it hits something"

        # Check if ethereal is to be disabled
        if self.ethereal and self.etherealTime + self.ETHEREAL_TIME < pygame.time.get_ticks():
            self.ethereal = 0

        # Check if floor is to be disabled
        if self.floor and self.floorTime + self.FLOOR_TIME < pygame.time.get_ticks():
            self.floor = 0

        # Ball launches
        if self.launched:

            # Hits bottom
            if self.rect.bottom > SCREENRECT.height + self.WIDTH:
                if self.floor == 0:
                    self.kill()
                else:
                    if self.direction == self.se:
                        self.direction = self.ne
                    else:
                        self.direction = self.nw

            # Hits roof
            elif self.rect.bottom < SCREENRECT.top + self.WIDTH:
                if self.direction == self.ne:
                    self.direction = self.se
                else:
                    self.direction = self.sw

            # Hits right side
            elif self.rect.right > SCREENRECT.right:
                if self.direction == self.ne:
                    self.direction = self.nw 
                else:
                    self.direction = self.sw

            # Hits left side
            elif self.rect.left < SCREENRECT.left:
                if self.direction == self.nw:
                    self.direction = self.ne
                else:
                    self.direction = self.se

            # Hits the bar
            elif self.rect.colliderect (bar.rect):
                bar.add (dirty_sprites)

                if self.direction == self.se:
                    self.direction = self.ne
                else:
                    self.direction = self.nw

                if bar.glue:
                    self.launch (0)
                    self.diffXpos = self.rect.centerx - bar.rect.centerx

            # Collides with block(s)
            blockshit = pygame.sprite.spritecollide (self, blocks, 0)
            if len (blockshit) > 0:
                for block in blockshit:
                    block.hit()

                centery = self.rect.centery
                top = block.rect.top
                bottom = block.rect.bottom

                # check if ball hits the side of a block
                hits_side = centery > top and centery < bottom

                if not self.ethereal:
                    if self.direction == self.nw:
                        if (hits_side):
                            self.direction = self.ne
                        else:
                            self.direction = self.sw
                    elif self.direction == self.ne:
                        if (hits_side):
                            self.direction = self.nw
                        else:
                            self.direction = self.se
                        
                    elif self.direction == self.sw:
                        if (hits_side):
                            self.direction = self.se
                        else:
                            self.direction = self.nw
                    else: # direction se
                        if (hits_side):
                            self.direction = self.sw
                        else:
                            self.direction = self.ne


            self.rect.move_ip (self.direction[0], self.direction[1])
        else:
            self.glue()

        self.add (moving_sprites)

class Block (pygame.sprite.Sprite):
    """Blocks the player has to shoot down"""

    def __init__ (self, x, y, kind):
        pygame.sprite.Sprite.__init__ (self)

        self.kind = kind
        self.type_setup ()
        self.rect = self.image.get_rect()
        self.rect.topleft = (x, y)

        self.updateGroup = pygame.sprite.RenderClear()
        self.add (self.updateGroup)
        self.draw()

    def draw (self):
        self.updateGroup.clear (screen.display, background)
        self.updateGroup.draw (screen.display)

    def type_setup (self):
        self.hits = 1
        self.points = 50

        if self.kind == '1': blockgfx = 'block1'
        elif self.kind == '2': blockgfx = 'block2'
        elif self.kind == '3': blockgfx = 'block3'
        elif self.kind == '4': blockgfx = 'block4'
        elif self.kind == '5': blockgfx = 'block5'
        elif self.kind == '6': blockgfx = 'block6'
        elif self.kind == '7':
            blockgfx = 'block7'
            self.hits = 3
            self.points = 25

        self.image = gfx[blockgfx]

    def hit (self):
        player.updateScore (self.points)
        self.hits -= 1

        if self.hits <= 0:
            self.kill()
            sfx['block_hit'].play()
            percentage = random.randrange (1, 100)
            if (percentage <= 25):
                Feature ((self.rect.centerx, self.rect.bottom)).add (features)
            self.draw()
        else:
            self.add (dirty_sprites)


class Shot (pygame.sprite.Sprite):
    """A nice bulleet"""

    SPEED = -2

    def __init__ (self):
        pygame.sprite.Sprite.__init__ (self)

        self.image = gfx['shot1']
        self.rect = self.image.get_rect()
        self.WIDTH = self.rect.width
        self.rect.centerx = bar.rect.centerx
        self.rect.centery = bar.rect.centery - self.WIDTH - 1
        sfx['shot'].play()

    def update (self):
        self.add (moving_sprites)

        self.rect.move_ip (0, self.SPEED)

        # Hits top 
        if self.rect.top < SCREENRECT.top:
            self.kill()

        # hits block
        blockshit = pygame.sprite.spritecollide (self, blocks, 0)
        if len (blockshit) > 0:
            for block in blockshit:
                block.hit()
            self.kill()


class Feature (pygame.sprite.Sprite):
    """Special goodie for the player"""

    def __init__ (self, topleft):
        pygame.sprite.Sprite.__init__ (self)

        self.name = random.choice (('a', 'b', 'c', 'd', 'e', 'f',
                                    'g', 'h', 'i', 'j', 'k', 'l'))
        self.speed = random.choice ((1, 2, 3, 4))        

        font = pygame.font.Font (None, 20)

        self.image = font.render (self.name, 1, (0, 0, 0))
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

        self.HEIGHT = self.rect.height

    def update (self):
        self.add (moving_sprites)

        # hits blocks
        blocksHit = pygame.sprite.spritecollide (self, blocks, 0)
        if (len (blocksHit) > 0):
            for block in blocksHit:
                block.add (dirty_sprites)

        self.rect.move_ip (0, self.speed)
                
        # hits bar
        if self.rect.colliderect (bar.rect):
            if self.name == 'l':
                for ball in balls.sprites():
                    ball.setSpeed (1)
            elif self.name == 'k':
                for ball in balls.sprites():
                    ball.setSpeed (-1)
            elif self.name == 'j':
                for ball in balls.sprites():
                    ball.setFloor()
            elif self.name == 'i':
                for ball in balls.sprites():
                    ball.setEthereal()
            elif self.name == 'h': bar.shots += 10
            elif self.name == 'g': bar.glue = 1
            elif self.name == 'f': Ball().add (balls)
            elif self.name == 'e': player.updateLives (1)
            elif self.name == 'd': player.updateScore (200)
            elif self.name == 'c': bar.setSpeed (1)
            elif self.name == 'b': bar.setSpeed (-1)
            elif self.name == 'a': player.updateLives (-1)
            
            self.kill()

        elif self.rect.bottom > SCREENRECT.height + self.HEIGHT:
            self.kill()

class Player:
    """A representation of a player"""

    def __init__ (self, name):
        self.name = name
        self.normalize()

    def normalize (self):
        self.lives = 3
        self.score = 0

    def updateLives (self, amount):
        self.lives += amount
        infobar.update()

    def updateScore (self, amount):
        self.score += amount
        infobar.update()

class InfoBar (pygame.sprite.Sprite):
    """An information bar containing vital info for the player"""

    def __init__ (self):
        pygame.sprite.Sprite.__init__ (self)

        self.font = pygame.font.Font(None, 16)
        self.image = self.font.render ('Level: 1\t Score: 0\t Lives: 3', 1, (255, 255, 255))
        self.rect = self.image.get_rect()
        self.rect.topleft = (5, 5)
        self.updateGroup = pygame.sprite.RenderClear()
        self.add (self.updateGroup)
        self.draw()

    def draw (self):
        self.updateGroup.clear (screen.display, background)
        self.updateGroup.draw (screen.display)

    def update (self):
        self.image = self.font.render (self._infoText(), 1, (255, 255, 255))
        self.draw()

    def _infoText (self):
        return '%s %i\t %s %i\t %s %i' % ('Level:', level.number, 'Score:', player.score,
                                                  'Lives:', player.lives)


class Level:
    """Spawns new levels"""
    number = 0

    def __init__ (self):
        self.BLOCK_WIDTH = gfx['block1'].get_rect().width
        self.BLOCK_HEIGHT = gfx['block1'].get_rect().height
        self.num_maps = len (os.listdir (os.path.join ('data', 'level')))
        self.new()
        
    def new (self):
        self.number += 1

        # load music
        intro_music[(self.number % 6) + 1].play()
        load_music (music[random.choice ((1,))])
        pygame.mixer.music.play (-1)

        # load blocks
        blocks.empty()

        # load map
        if (self.number <= self.num_maps):
            map = self.load_map (self.number)
        else:
            map = self.random_map()
    
        yPos = 30
        for y in range (len (map)):
        
            xPos = 75
            for x in map[y]:
                if x != '0':
                    block = Block (xPos, yPos, x)
                    block.add (blocks)


                xPos += self.BLOCK_WIDTH
                
            yPos += self.BLOCK_HEIGHT
            

    def random_map (self):
        map = []

        for y in range (10):
            innermap = []
            for x in range (12):
                innermap.append (random.choice (('0', '1', '2', '3',
                                                 '4', '5', '6', '7')))

            map.append (innermap)

        return map

    def load_map (self, num):
        map = []
        path = os.path.join ('data', 'level', str (num))
        fp = open (path)
        lines = fp.readlines()

        for line in lines:
            innermap = []
            for num in line:
                if num != '\n':
                    innermap.append (num)
            map.append (innermap)

        fp.close()
        return map

class Hiscore:
    """A nice hiscore list"""

    def __init__ (self, filename):
        self.filename = filename
        self.text = self.read()
        self.font = fonts['std']

    def draw (self):
        image = pygame.Surface ((500, 300)).convert()
        image.fill ((91, 159, 211))

        y = 0
        for line in self.text[0]:
            pos = (0, y)
            lineImage = self.font.render (line, 1, (255, 255, 255))
            image.blit (lineImage, pos)
            y += 30

        y = 0
        for line in self.text[1]:
            pos = (350, y)
            lineImage = self.font.render (line, 1, (255, 255, 255))
            image.blit (lineImage, pos)
            y += 30

        y = 0
        for line in self.text[2]:
            pos = (425, y)
            lineImage = self.font.render (line, 1, (255, 255, 255))
            image.blit (lineImage, pos)
            y += 30


        rect = image.get_rect()
        rect.center = SCREENRECT.center

        screen.display.blit (gfx['bg2'], (0, 0))
        screen.display.blit (image, rect)
        pygame.display.flip()

    def read (self):
        file = ['Name:Level:Score',
                'Martin Willemoes Hansen:5:10000',
                'foobar1:4:8000',
                'foobar2:3:6000',
                'foobar3:2:5000',
                'foobar4:2:4500',
                'foobar5:1:3000',
                'foobar6:1:2000',
                'foobar7:1:1500',
                'foobar8:1:1000',
                'foobar9:1:500']

        output = [[], [], []]

        for line in file:
            elements = line.split (':')
            output[0].append (elements[0])
            output[1].append (elements[1])
            output[2].append (elements[2])

        return output
                
    
    def save (self):
        print 'save hiscore to file'


# functions
def game_over():
    player.normalize()

    text = ['Game over!!!',
            'Hiscore [h]',
            'New game [space]',
            'Quit [ESC]']

    textEntries = len (text) * 30

    font = fonts['std']
    screen.display.blit (gfx['bg2'], (0,0))
    y = 0
    for line in text:
        textImage = font.render (line, 1, (255, 255, 255))
        textRect = textImage.get_rect()
        textRect.centerx = SCREENRECT.centerx
        textRect.centery = SCREENRECT.centery + y - textEntries

        screen.display.blit (textImage, textRect)
        if y == 0:
            y = 30
        y += 30

    pygame.display.flip()

    quit = 1 # dont quit before esc clicked again
    hiscoreSet = 0

    while 1:
        pygame.event.pump()
        keyinput = pygame.key.get_pressed()
    
        if not quit and keyinput[K_ESCAPE] or pygame.event.peek(QUIT):
            credits()
            pygame.time.delay (3000)
            print ('Game over')
            sys.exit()
        quit = keyinput[K_ESCAPE]

        if keyinput[K_h]:
            hiscore.draw()
            pygame.time.delay (5000)
            hiscoreSet = 1
            break

        if keyinput[K_SPACE]:
            break

    if hiscoreSet:
        game_over()
        

def credits():
    text = ('Coding:\t Martin Willemoes Hansen aka MWH',
            'Music:\t\t DTC')

    font = fonts['std']

    images = []
    for line in text:
        images.append (font.render (line, 1, (255, 255, 255)))

    imageHeight = images[0].get_rect().height
    imagesNum = len (images)

    textImage = pygame.Surface ((465, imageHeight * imagesNum + 100)).convert()
    textImage.fill ((91, 159, 211))
    y = 0
    for image in images:
        textImage.blit (image, (0, y))
        y += imageHeight + 10
    
    textRect = textImage.get_rect()
    textRect.center = SCREENRECT.center

    screen.display.blit (gfx['bg2'], (0,0))
    screen.display.blit (textImage, textRect)
    pygame.display.flip()


def create_first_ball():
    balls.empty()
    firstBall = Ball()
    firstBall.launch(0)
    firstBall.add (balls)

def main():
    #Should make python interpret faster
    sys.setcheckinterval (100)

    pygame.init()
    pygame.mouse.set_visible (0)

    global screen, player, clock
    screen = Screen (640, 480, '-=* Pykanoid *=-')
    player = Player ('foobar')
    clock = pygame.time.Clock()

    load_files()

    #sprites and groups
    global balls, blocks, shots, features, moving_sprites, dirty_sprites
    global bar, hiscore, background

    background = gfx['bg2']

    balls = pygame.sprite.RenderClear()
    blocks = pygame.sprite.RenderClear()
    shots = pygame.sprite.RenderClear()
    features = pygame.sprite.RenderClear()
    moving_sprites = pygame.sprite.RenderClear()
    dirty_sprites = pygame.sprite.RenderPlain()

    bar = Bar()

    hiscore = Hiscore ('hiscore')

    game_loop()

def load_files():
    background = pygame.Surface (screen.display.get_size()).convert()
    background.fill ((91, 159, 211))

    global gfx
    gfx = {'bg1': load_image ('background1.jpeg'),
           'bg2': background,
           'ball': load_image ('ball.gif'),
           'bar1': load_image ('bar.gif'),
           'block1': load_image ('block1.jpeg'),
           'block2': load_image ('block2.jpeg'),
           'block3': load_image ('block3.jpeg'),
           'block4': load_image ('block4.jpeg'),
           'block5': load_image ('block5.jpeg'),
           'block6': load_image ('block6.jpeg'),
           'block7': load_image ('block7.jpeg'),
           'shot1': load_image ('shot.gif')}

    global music
    music = {1: 'anthem.mod'}

    global intro_music
    intro_music = {1: load_sound ('im_1.wav'),
                   2: load_sound ('im_2.wav'),
                   3: load_sound ('im_3.wav'),
                   4: load_sound ('im_4.wav'),
                   5: load_sound ('im_5.wav'),
                   6: load_sound ('im_6.wav')}

    global sfx
    sfx = {'shot': load_sound ('shot.wav'),
           'block_hit': load_sound ('block_hit.wav')}

    global fonts
    fonts = {'std': load_font (28)}

def reset():
    bar.normalize()
    bar.draw()
    create_first_ball()

def game_loop():
    screen.display.blit (background, (0,0))
    pause = 0
    reset()

    global infobar, level
    infobar = InfoBar()
    level = Level()

    while 1:
        clock.tick (150)
        pygame.event.pump()
        keyinput = pygame.key.get_pressed()

        # tired of game
        if keyinput[K_ESCAPE] or pygame.event.peek(QUIT):
            game_over()
            break

        # pause
        pauseNu = keyinput[K_p]
        if not pause and pauseNu:
            pauseInner = 1
            while 1:
                pygame.event.pump()
                keyinput = pygame.key.get_pressed()
                
                pauseInnerNu = keyinput[K_p]
                if not pauseInner and pauseInnerNu:
                    break
                pauseInner = pauseInnerNu
        pause = pauseNu

        # Movement of bar
        if keyinput[K_RIGHT]:
            bar.move (bar.RIGHT)
            
        if keyinput[K_LEFT]:
            bar.move (bar.LEFT)

        # catapults the ball at start of level
        if keyinput[K_SPACE]:
            for ball in balls.sprites():
                ball.launch (1)

        # determin if player has more lives
        if len (balls) == 0:
            player.updateLives (-1)

            if player.lives <= 0:
                game_over()
                break
            else:
                reset()

        # should we go to next level
        if len (blocks) == 0:
            moving_sprites.clear (screen.display, background)
            moving_sprites.empty()
            reset()
            level.new()
            infobar.update()

        # bar shoots
        firing = keyinput[K_SPACE]
        if not bar.reloading and firing and bar.shots > 0:
            bar.shots -= 1
            shot = Shot()
            shot.add (shots)
        bar.reloading = firing

        # update and draw
        balls.update()
        shots.update()
        features.update()

        moving_sprites.clear (screen.display, background)
        moving_sprites.draw (screen.display)
        moving_sprites.empty()

        dirty_sprites.draw (screen.display)
        dirty_sprites.empty()

        pygame.display.flip()


    game_loop()

if __name__ == '__main__':
    main()
